#include "base.h"
#include <iostream>
#include "muh_blood.h"
#include "melanin.h"
using namespace std;


void GPUSkinBase::allocate(int samples, int bands, size_t byteWidth, int height, const vector<float> &wavelengths){
	int arrayByteSize = sizeof(float)*samples*bands;
	size_t pitch;
	
	//wavelength dependencies of scattering and absorption coefficients
	cerr << cudaMallocPitch(&muh_oxy, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&muh_deoxy, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&melanin_base, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&musm, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&musr, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&musb_base, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&gcol, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&(this->wavelengths), &pitch, byteWidth, height);

	//calculate the spectrum arrays on the host, try to transfer concurrently by swapping between initiating the cudaMemCpy and doing calculations on the CPUi
	float *muh_oxy_host = (float*)malloc(arrayByteSize);
	float *muh_deoxy_host = (float*)malloc(arrayByteSize);
	float *melanin_base_host = (float*)malloc(arrayByteSize);
	float *gcol_host = (float*)malloc(arrayByteSize);
	float *musm_host = (float*)malloc(arrayByteSize);
	float *musr_host = (float*)malloc(arrayByteSize);
	float *musb_base_host = (float*)malloc(arrayByteSize);
	float *wavelengths_host = (float*)malloc(arrayByteSize);

	//prepare the wavelength dependent parts of the absorption and scattering coefficients
	//it might seem inefficient to assign the exact same value to the whole range of spatial positions for each wavelength, but even if it is all a memory waste, it will pay off in speed later on because of memory coalescing in the GPU. Also tried to load it into shared memory across all the thread blocks, but had to make the threads diverge in order to do that and would not work well with multiple shared variables
	//Would be the exact same latency issues. whether 32 threads load 32 different variables in one go and in parallell, or one of the thread loads the whole thing into the cache and picks one variable and broadcasts it to the rest will give the exact same latency, plus it will have to stall threads. 
	float lambda;
	int position;
	float muh_oxy_temp, muh_deoxy_temp, gcol_temp, musm_temp, musr_temp, musb_temp, melanin_temp;
	int spatial_size = samples;
	for (int i=0; i < bands; i++){
		lambda = wavelengths[i];
		muh_oxy_temp = muh_oxy_calc(lambda);
		muh_deoxy_temp = muh_deoxy_calc(lambda);
		gcol_temp = 0.62 + lambda*29e-5;
		musm_temp = (1-1.745e-3*lambda + 9.843e-7*lambda*lambda)/(1-gcol_temp);
		musr_temp = pow(lambda, -4);
		musb_temp = pow((685/(lambda*1.0)), 0.37);
		melanin_temp = melanin(lambda);
		for (int j=0; j < samples; j++){
			position = i*spatial_size + j;
			muh_oxy_host[position] = muh_oxy_temp;
			muh_deoxy_host[position] = muh_deoxy_temp;
			melanin_base_host[position] = melanin_temp;
			gcol_host[position] = gcol_temp;
			musm_host[position] = musm_temp;
			musr_host[position] = musr_temp;
			musb_base_host[position] = musb_temp/(1-gcol_temp)*(1-0.996);
		}
		wavelengths_host[i] = lambda;

	}

	//copy the wavelength dependencies to the GPU device
	cerr << cudaMemcpy2D(muh_oxy, pitch, muh_oxy_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(muh_deoxy, pitch, muh_deoxy_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(melanin_base, pitch, melanin_base_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(musm, pitch, musm_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(musr, pitch, musr_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(musb_base, pitch, musb_base_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(gcol, pitch, gcol_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);
	cerr << cudaMemcpy2D(this->wavelengths, pitch, wavelengths_host, byteWidth, byteWidth, height, cudaMemcpyHostToDevice);

	//free arrays that are no longer needed
	free(muh_oxy_host);
	free(muh_deoxy_host);
	free(melanin_base_host);
	free(musm_host);
	free(musr_host);
	free(musb_base_host);
	free(gcol_host);
	free(wavelengths_host);
}

void GPUSkinData::allocate(size_t byteWidth, int height, int thrPerB){
	cerr << cudaMallocPitch(&(oxy), &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&(bvf), &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&(muam694), &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&(melanin_type), &pitch, byteWidth, height);
	this->byteWidth = byteWidth;
	this->height = height;
	this->pitch = pitch;
}

void GPUOpticalProps::allocate(size_t byteWidth, int height){
	size_t pitch;
	cerr << cudaMallocPitch(&muae, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&muse, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&muad, &pitch, byteWidth, height);
	cerr << cudaMallocPitch(&musd, &pitch, byteWidth, height);
}

float *GPUSkinData::download(int samples){

	float *muamhost = new float[samples];
	float *bvfhost = new float[samples];
	float *oxyhost = new float[samples];
	cerr << cudaMemcpy2D(muamhost, byteWidth, muam694, pitch, byteWidth, height, cudaMemcpyDeviceToHost);
	cerr << cudaMemcpy2D(bvfhost, byteWidth, bvf, pitch, byteWidth, height, cudaMemcpyDeviceToHost);
	cerr << cudaMemcpy2D(oxyhost, byteWidth, oxy, pitch, byteWidth, height, cudaMemcpyDeviceToHost);
	
	/*for (int i=0; i < samples; i++){
		cout << muamhost[i] << " ";
	}
	cout << endl;*/

	float *retarr = new float[2*samples];

	memcpy(retarr, bvfhost, samples*sizeof(float));
	memcpy(retarr+samples, oxyhost, samples*sizeof(float));

	delete [] muamhost;
	delete [] bvfhost;
	delete [] oxyhost;

	return retarr;
}

