#include "fitting.h"
#include "../CPU-DM/nnls.h"
#include <gsl/gsl_vector.h>
#include "muh_blood.h"
#include <cmath>
#include <hyLight/FloatImage.h>
#include "math.h"
#include "hansonfitting.h"
#include <fstream>
using namespace std;


Fitting::Fitting(Chromophores chrom, const vector<float> &wlens, float startwlen, float endwlen, int samples, int bands) : _samples(samples), _wlens(wlens), _bands(bands), _startwlen(startwlen), _antWlens(endwlen-startwlen), _chrom(chrom){
	generateInterpolTables();
	_endmembers = chrom.getNumEndmembers();
}


void Fitting::generateInterpolTables(){
	//start and end wavelengths in interpolation
	_minWlen = (int)(_wlens[0]+1);
	_maxWlen = (int)(_wlens.back());
	_stepWlen = 1;

	//lists of interpolation indices
	int antwlens = (_maxWlen - _minWlen)/_stepWlen;
	_uppers = new int[antwlens];
	_downers = new int[antwlens];

	//go through the list of wavelengths and find the interpolation wavelengths
	float currwlen = _minWlen;
	int band = 0;
	int ind = 0;
	while (band < _bands-1){
		float w = _wlens[band];
		float wn = _wlens[band+1];
		if (((w <= currwlen) && (wn > currwlen)) && (ind < antwlens)){
			_downers[ind] = band;
			_uppers[ind] = band+1;
			ind++;
			currwlen += _stepWlen;
		} else {
			band++;
		}
	}

	//save the wavelengths so we don't have to look them up
	_upperWlens = new float[antwlens];
	_downerWlens = new float[antwlens];
	for (float w = _minWlen; w < _maxWlen; w += _stepWlen){
		int ind = w - _minWlen;
		_downerWlens[ind] = _wlens[_downers[ind]];
		_upperWlens[ind] = _wlens[_uppers[ind]];
	}
}


float** Fitting::generateChromophoreMatrix(Chromophores chrom, float startwlen, int antWlens, int endmembers){
	float **retMat = new float*[antWlens];
	for (int i=0; i < antWlens; i++){
		float wlen = startwlen + i;
		retMat[i] = chrom.getAbsArray(wlen);
	}
	return retMat;
}
		
		
HansonFitting::HansonFitting(Chromophores chrom, const vector<float> &wlens, float startwlen, float endwlen, int samples, int bands) : Fitting(chrom, wlens, startwlen, endwlen, samples, bands){
	generateHansonMatrix();
	createPorts();
}
void HansonFitting::generateHansonMatrix(){
	//use the fitting parent class to create the chromophore matrix, move it to an gsl array
	float **chromMat = Fitting::generateChromophoreMatrix(_chrom, _startwlen, _antWlens, _endmembers);
	_chromMat = gsl_matrix_alloc(_antWlens, _endmembers);
	for (int i=0; i < _antWlens; i++){
		for (int j=0; j < _endmembers; j++){
			gsl_matrix_set(_chromMat, i, j, chromMat[i][j]);
		}
	}
	for (int i=0; i < _antWlens; i++){
		delete [] chromMat[i];
	}
	delete [] chromMat;
}


ofstream oxyut;
ofstream bvfut;
float *HansonFitting::doHansonFitting(float *mua){
	float *res = new float[_samples*_endmembers];

	int startpix = 0;
	int endpix = _samples;

	//go through the pixels, invert each separatedly
	//#pragma omp parallel for
	for (int pix = 0; pix < _samples; pix++){
		gsl_vector *abs = gsl_vector_alloc(_antWlens);
		gsl_vector *c = gsl_vector_alloc(_endmembers);

		//only invert selected pixels
		if ((pix >= startpix) && (pix <= endpix)){
			//interpolate like a motherfucking north star
			for (int w = _startwlen; w < _startwlen+_antWlens; w++){
				int i=w-_minWlen;
				float lowabs = mua[_samples*_downers[i] + pix];
				float uppabs = mua[_samples*_uppers[i] + pix];
				float lower_l = _wlens[_downers[i]];
				float upper_l = _wlens[_uppers[i]];
				gsl_vector_set(abs, w-_startwlen, (lowabs + (w - lower_l)/(upper_l - lower_l)*(uppabs - lowabs)));
			}

			double *std = new double[_endmembers];

			//fitting
			nnls_hanson(_chromMat, abs, _endmembers, _antWlens, c, std);

			//save the results
			res[0*_samples + pix] = gsl_vector_get(c, 0);//gsl_vector_get(c, 0) + gsl_vector_get(c, 1);
			res[1*_samples + pix] = gsl_vector_get(c, 1);///res[0*_samples + pix];
			for (int i=2; i < _endmembers; i++){
				res[i*_samples + pix] = gsl_vector_get(c, i);
			}
			delete [] std;
		} else {
			for (int i=0; i < _endmembers; i++){
				res[i*_samples + pix] = 0;
			}
		}
		bvfut << gsl_vector_get(c,0) + gsl_vector_get(c,1) << " ";
		oxyut << gsl_vector_get(c,0)/(gsl_vector_get(c,0) + gsl_vector_get(c,1)) << " ";

		gsl_vector_free(abs);
		gsl_vector_free(c);
	}
	bvfut << endl;
	oxyut << endl;
	return res;
}
void HansonFitting::execute(){
	//omitted
}
