//helper functions for common frame of reference for what indices should correspond to what wavelengths
int getStartInd(vector<float> wlens, float startwlen){
	//find the wavelength indices corresponding to the specificed wavelength interval
	int startwlenInd;
	bool foundStart = false;
	for (unsigned int i=0; i < wlens.size(); i++){
		if ((wlens[i] > startwlen) && (!foundStart)){
			startwlenInd = i-1;
			foundStart = true;
			break;
		}
	}
	if (!foundStart){
		startwlenInd = 0;
	}
	return startwlenInd;
}
int getEndInd(vector<float> wlens, float endwlen){
	//find the wavelength indices corresponding to the specificed wavelength interval
	int endwlenInd;
	bool foundEnd = false;
	for (unsigned int i=0; i < wlens.size(); i++){
		if ((wlens[i] > endwlen) && (!foundEnd)){
			endwlenInd = i-1;
			foundEnd = true;
			break;
		}
	}
	if (!foundEnd){
		endwlenInd = wlens.size()-1;
	}
	return endwlenInd;
}

//GPU-DM
GPUDM::GPUDM(int bands, int samples, const vector<float> &wlens) : _bands(bands), _samples(samples), _wlens(wlens){
	//start and end wavelengths for melanin determination
	float startwlen = 730;
	float endwlen = 830;
	
	num_melanin_iterations = 2;

	//GPU array and calc properties
	cudaStreamCreate(&transferStream);
	cudaStreamCreate(&processingStream);
	_threadsPerBlock = 160;

	_byteWidth = _threadsPerBlock*sizeof(float);
	_height = _samples*_bands/_threadsPerBlock;
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = _bands;
	_dimBlock = dim3(_threadsPerBlock);
	_dimGrid = dim3(block_number_x, block_number_y);

	//create buffers
	cudaMallocPitch(&_currRefl, &_pitch, _byteWidth, _height);
	
	//dummy array for inclusion of methb in first SCA
	methbDummyInd = 0;
	cudaMallocPitch(&methbdummy, &_pitch, sizeof(float)*_threadsPerBlock, _samples/_threadsPerBlock);
	float *dummy = new float[_samples];
	for (int i=0; i < _samples; i++){
		dummy[i] = 1000;
	}
	cudaMemcpy2D(methbdummy, _pitch, dummy, sizeof(float)*_threadsPerBlock, sizeof(float)*_threadsPerBlock, _samples/_threadsPerBlock, cudaMemcpyHostToDevice);
	delete [] dummy;

	//inversion arrays 
	_skinBase = new GPUSkinBase;
	_skinBase->allocate(_samples, _bands, _byteWidth, _height, wlens);
	_skinDataMuadInv = new GPUSkinData;
	_skinDataMuadInv->allocate(_byteWidth, _height, _threadsPerBlock);
	_skinDataMuaeInv = new GPUSkinData;
	_skinDataMuaeInv->allocate(_byteWidth, _height, _threadsPerBlock);

	_optProps = new GPUOpticalProps;
	_optProps->allocate(_byteWidth, _height);
	cudaMallocPitch(&_simrefl, &_pitch, _byteWidth, _height);
	cudaMallocPitch(&_deriv, &_pitch, _byteWidth, _height);
	_numIterations = 15;

	//find the wavelength indices corresponding to the specificed wavelength interval
	_startwlenInd = getStartInd(_wlens, startwlen);
	_endwlenInd = getEndInd(_wlens, endwlen);
	
	//initialize skin data	
	float *muamtemp = new float[_samples];
	float *oxytemp = new float[_samples];
	float *bvftemp = new float[_samples];
	float *melanintypetemp = new float[_samples];

	for (int i=0; i < _samples; i++){
		muamtemp[i] = 100;
		oxytemp[i] = 0.8;
		bvftemp[i] = 0.01;
		melanintypetemp[i] = SVAASAND_MELANIN_GPU;
	}
	size_t SDByteWidth = sizeof(float)*_threadsPerBlock;
	int SDHeight = _samples/_threadsPerBlock;
	
	cudaMemcpy2D(_skinDataMuadInv->muam694, _pitch, muamtemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuadInv->melanin_type, _pitch, melanintypetemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuadInv->oxy, _pitch, oxytemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuadInv->bvf, _pitch, bvftemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuaeInv->muam694, _pitch, muamtemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuaeInv->melanin_type, _pitch, melanintypetemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuaeInv->oxy, _pitch, oxytemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);
	cudaMemcpy2D(_skinDataMuaeInv->bvf, _pitch, bvftemp, SDByteWidth, SDByteWidth, SDHeight, cudaMemcpyHostToDevice);

	delete [] muamtemp;
	delete [] oxytemp;
	delete [] bvftemp;

	//melunmixreadyment
	float *melcurveHost = new float[_samples*_bands];
	_factor = 0;
	for (int i=_startwlenInd; i < _endwlenInd; i++){
		float temp = pow(694.0f/wlens[i], 3.46);
		_factor += temp*temp;
		for (int j=0; j < _samples; j++){
			melcurveHost[i*_samples + j] = temp;
		}
	}
	_factor = 1.0f/_factor;
	cudaMallocPitch(&_melcurve, &_pitch, _byteWidth, _height);
	cudaMemcpy2D(_melcurve, _pitch, melcurveHost, _byteWidth, _byteWidth, _height, cudaMemcpyHostToDevice);
	delete [] melcurveHost;

	//fitting classes and factory classes
	Chromophores chrom;
	chrom.setMel();

	float melstart = startwlen + 10;
	float melend = endwlen - 10;


	chrom.unsetMel();
	chrom.setKonst();
	

	Chromophores melchrom;
	melchrom.setMel();
	melchrom.setWat();
	scabefore = new SCAFitting(melchrom, wlens, startwlen, endwlen, _samples, _bands, _threadsPerBlock, _pitch);

	chrom.setMel();
	chrom.setMethb();
	chrom.setBil();
	chrom.setBet();
	
	sca450 = new SCAFitting(chrom, wlens, 460, 530, _samples, _bands, _threadsPerBlock, _pitch);
	objFact1 = new ChromImageFactory(_samples, sca450->getNumEndmembers());

	chrom.unsetBil();
	chrom.unsetBet();
	
	sca530 = new SCAFitting(chrom, wlens, 500, 590, _samples, _bands, _threadsPerBlock, _pitch);
	objFact2 = new ChromImageFactory(_samples, sca530->getNumEndmembers());
	
	//620-730
	//690-820 er ogsaa ok
	sca700 = new SCAFitting(chrom, wlens, 620, 730, _samples, _bands, _threadsPerBlock, _pitch);
	objFact3 = new ChromImageFactory(_samples, sca700->getNumEndmembers());
	
	melFact = new ChromImageFactory(_samples, 1);
	meltypeFact = new ChromImageFactory(_samples, 1);
	errorFact = new ChromImageFactory(_samples, 2);
	muadFact = new ChromImageFactory(_samples, _bands);
	createPorts();


	cudaMallocPitch(&_gpumuad, &_pitch, _byteWidth, _height);
	cudaMallocPitch(&_gpures, &_pitch, _byteWidth, _height);
	cudaMallocPitch(&_gpureflscaling, &_pitch, _byteWidth, _height);

	//wavelength indices for calculating absorption scaling
	_scalewlenind1 = getStartInd(_wlens, SCALE_WLEN_1);
	_scalewlenind2 = getStartInd(_wlens, SCALE_WLEN_2);
}
	

void GPUDM::checkReflectanceScaling(float *refl, GPUSkinData *skinData, int wlenind1, int wlenind2, dim3 dimGrid, size_t inputPitch){
	//invert at wavelength 1
	calcSkinData<<<dimGrid, _dimBlock>>>(_skinBase->wavelengths, skinData->oxy, skinData->bvf, skinData->muam694, skinData->melanin_type, _optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->muh_oxy, _skinBase->muh_deoxy, _skinBase->melanin_base, _skinBase->musm, _skinBase->musr, _skinBase->musb_base, inputPitch, wlenind1);
	ReflIsoL2InvertMuad<<<dimGrid, _dimBlock>>>(_optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->gcol, refl, inputPitch, wlenind1);

	//invert at wavelength 2
	calcSkinData<<<dimGrid, _dimBlock>>>(_skinBase->wavelengths, skinData->oxy, skinData->bvf, skinData->muam694, skinData->melanin_type, _optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->muh_oxy, _skinBase->muh_deoxy, _skinBase->melanin_base, _skinBase->musm, _skinBase->musr, _skinBase->musb_base, inputPitch, wlenind2);
	ReflIsoL2InvertMuad<<<dimGrid, _dimBlock>>>(_optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->gcol, refl, inputPitch, wlenind2);

	//check the muad scaling, decide whether to move the melanin type to the left or not
	CheckReflectanceScaling<<<dimGrid, _dimBlock>>>(_optProps->muad, skinData->melanin_type, wlenind1, wlenind2, inputPitch, SCALING_THRESHOLD_MELANINTYE_DETECTION);
}

void GPUDM::detectMelaninType(float *refl, GPUSkinData *skinData){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	dim3 dimGrid = dim3(block_number_x, block_number_y);
	size_t inputPitch = _pitch/sizeof(float);

	//set the melanin type to PHEOMELANIN_GPU (is originally: svaasand)
	SetMelaninType<<<dimGrid, _dimBlock>>>(skinData->melanin_type, PHEOMELANIN_GPU, inputPitch);

	//check whether we should move up from PHEOMELANIN_GPU to SVAASAND_MELANIN_GPU 
	checkReflectanceScaling(refl, skinData, _scalewlenind1, _scalewlenind2, dimGrid, inputPitch);

	//check whether we should move up from SVAASAND_MELANIN_GPU to EUMELANIN_GPU
	checkReflectanceScaling(refl, skinData, _scalewlenind1, _scalewlenind2, dimGrid, inputPitch);

	//check whether it still makes no sense
	checkReflectanceScaling(refl, skinData, _scalewlenind1, _scalewlenind2, dimGrid, inputPitch);

	//congratulations, the correct melanin type has now hopefully been detected!
}

//inversion of reflectance for one hyperspectral line of data
void GPUDM::invertReflectance(float *refl, bool printToFile, int pixel){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	dim3 dimGrid = dim3(block_number_x, block_number_y);
	size_t inputPitch = _pitch/sizeof(float);
	skindataDeinit<<<dimGrid, _dimBlock, 0>>>(_skinDataMuaeInv->muam694, _skinDataMuaeInv->bvf, _skinDataMuaeInv->oxy, _skinDataMuaeInv->melanin_type, inputPitch);

	//upload to GPU
	cudaMemcpy2D(_currRefl, _pitch, refl, _byteWidth, _byteWidth, _height, cudaMemcpyHostToDevice);

	//if we should print to file: print chosen reflectance to file
	if (printToFile){	
		//omitted for readability
	}
	
	//find required melanin
	invertMuam694(num_melanin_iterations, printToFile, pixel);
	
	//attempt melanin type detection
	detectMelaninType(_currRefl, _skinDataMuaeInv);

	//invert muad
	invertMuad(_skinDataMuaeInv, _simrefl, _deriv, _currRefl, 0, _bands-1, &processingStream);

	//do fitting at the pre-specified wavelength ranges
	sca700->doFitting(_skinDataMuaeInv, _optProps->muad, false, false, methbdummy, methbDummyInd);
	sca450->doFitting(_skinDataMuaeInv, _optProps->muad, false, false, sca700->getRes(), sca700->getMethbInd());
	sca530->doFitting(_skinDataMuaeInv, _optProps->muad, false, false, sca700->getRes(), sca700->getMethbInd());

	//if we are to print to file, print muad of selected pixel, calculate forward reflectance of selected pixel
	if (printToFile){
		//omitted for readability
	}
}

//find muam694
void GPUDM::invertMuam694(int iterations, bool printToFile, int pixel){
	int num_muae = _endwlenInd-_startwlenInd;
	int onechrStart = _startwlenInd;
	int onechrEnd = _endwlenInd;

	for (int i=0; i < iterations; i++){
		//invert muad
		invertMuad(_skinDataMuaeInv, _simrefl, _deriv, _currRefl, _startwlenInd, _endwlenInd, &processingStream);
		float *muadnocorr;
		if (printToFile){
			//omitted for readability
		}

		//fit the absorption spectrum
		scabefore->doFitting(_skinDataMuaeInv, _optProps->muad, true, true, methbdummy, methbDummyInd);

		//intermediate printing to file if so desired
		if (printToFile){
			//omitted for readability
		}

		//choose how to determine the melanin: straight line fitting of epidermis or melanin curve fitting of epidermis
		MelaninDetermination melDet;
		if (i < iterations-1){
			melDet = USE_STRAIGHT_LINE;
		} else {
			melDet = USE_MELANIN_CURVE;
		}

		//invert muae
		invertMuae(_skinDataMuaeInv, _simrefl, _deriv, _currRefl, _startwlenInd, _startwlenInd + num_muae, &processingStream, false);

		//use either straight line fitting or melanin curve and fit
		doOneChromSpectralUnmixing(_skinDataMuaeInv, _optProps->muae, onechrStart, onechrEnd, &processingStream, melDet);

		//print to file if required
		if (indivprint){
			//omitted for readability
		}
	}
}

void GPUDM::execute(){
	//omitted
}

//invert respectively muad and muae given data and input parameters
void GPUDM::invertMuad(GPUSkinData *skinData, float *simrefl, float *deriv, float *currRefl, int startwlenInd, int endwlenInd, cudaStream_t *processingStream){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = endwlenInd - startwlenInd;
	
	dim3 dimGrid = dim3(block_number_x, block_number_y);

	size_t inputPitch = _pitch/sizeof(float);
	calcSkinData<<<dimGrid, _dimBlock, 0, *processingStream>>>(_skinBase->wavelengths, skinData->oxy, skinData->bvf, skinData->muam694, skinData->melanin_type, _optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->muh_oxy, _skinBase->muh_deoxy, _skinBase->melanin_base, _skinBase->musm, _skinBase->musr, _skinBase->musb_base, inputPitch, startwlenInd);


	ReflIsoL2InvertMuad<<<dimGrid, _dimBlock, 0, *processingStream>>>(_optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->gcol, currRefl, inputPitch, startwlenInd);
			
	
}	
void GPUDM::invertMuae(GPUSkinData *skinData, float *simrefl, float *deriv, float *currRefl, int startwlenInd, int endwlenInd, cudaStream_t *processingStream, bool shouldCalcSkinData){
	//NB FIXME: sykt udynamisk maate aa opprette griddet paa, burde tilpasse seg den som allerede er til stede.
	int threadsPerBlock = 64;
	int block_number_x = _samples/threadsPerBlock;
	int block_number_y = endwlenInd - startwlenInd;
	
	dim3 dimBlockRefl = dim3(threadsPerBlock);
	dim3 dimGridRefl = dim3(block_number_x, block_number_y);

	block_number_x = _samples/_threadsPerBlock;
	dim3 dimGridCalc = dim3(block_number_x, block_number_y);
	dim3 dimBlockCalc = dim3(_threadsPerBlock);

	size_t inputPitch = _pitch/sizeof(float);
	if (shouldCalcSkinData){
		calcSkinData<<<dimGridCalc, dimBlockCalc, 0, *processingStream>>>(_skinBase->wavelengths, skinData->oxy, skinData->bvf, skinData->muam694, skinData->melanin_type, _optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->muh_oxy, _skinBase->muh_deoxy, _skinBase->melanin_base, _skinBase->musm, _skinBase->musr, _skinBase->musb_base, inputPitch, startwlenInd);
	}

	//FIXME: _currRefl??? hvorfor medlemsvariabel
	ReflIsoL2InvertMuae<<<dimGridRefl, dimBlockRefl, 0, *processingStream>>>(_optProps->muae, _optProps->muse, _optProps->muad, _optProps->musd, _skinBase->gcol, _currRefl, inputPitch, startwlenInd);
	
}


void GPUDM::doOneChromSpectralUnmixing(GPUSkinData *skinData, float *gpuMua, int startwlenInd, int endwlenInd, cudaStream_t *processingStream, MelaninDetermination choice){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	size_t inputPitch = _pitch/sizeof(float);

	dim3 dimGrid(block_number_x, block_number_y);
	dim3 dimBlock(_threadsPerBlock);
	
	if (choice == USE_STRAIGHT_LINE){
		//approximate the epidermal absorption to a straight line
		StraightLine<<<dimGrid, dimBlock, 0, *processingStream>>>(_skinBase->wavelengths, gpuMua, MELANIN_REFERENCE_WAVELENGTH, skinData->muam694, startwlenInd, endwlenInd, inputPitch);
	} else if (choice == USE_MELANIN_CURVE){
		//use the melanin curve
		MultVector<<<dimGrid, dimBlock, 0, *processingStream>>>(_melcurve, gpuMua, skinData->muam694, _factor, startwlenInd, endwlenInd, inputPitch);
	}

}

ISRAFitting::ISRAFitting(Chromophores chrom, const vector<float> &wlens, float startwlen, float endwlen, int samples, int bands, int threadsPerBlock, size_t pitch) : SCAFitting(chrom, wlens, startwlen, endwlen, samples, bands, threadsPerBlock, pitch){
}

void ISRAFitting::doFitting(GPUSkinData *t, float *gpuMua){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	dim3 dimBlock(_threadsPerBlock);
	dim3 dimGrid(block_number_x, block_number_y);
	size_t inputPitch = _pitch/sizeof(float);
	ISRA<160><<<dimGrid, dimBlock, 0>>>(gpuMua, _res, _GPUA, inputPitch, _startind, _endmembers, _lseantWlens);
}

SCAFitting::SCAFitting(Chromophores chrom, const std::vector<float> &wlens, float startwlen, float endwlen, int samples, int bands, int threadsPerBlock, size_t pitch){
	_chrom = chrom;
	_wlens = wlens;
	_samples = samples;
	_bands = bands;
	_threadsPerBlock = threadsPerBlock;
	_pitch = pitch;
	_endmembers = chrom.getNumEndmembers();
	_byteWidth = sizeof(float)*_threadsPerBlock;
	_currMethbInd = chrom.getMethbInd();
	_startind = getStartInd(wlens, startwlen);
	_endind = getEndInd(wlens, endwlen);
	_lseantWlens = _endind - _startind;
	_lsestartwlen = wlens[_startind];
	prepSCA();
}

void SCAFitting::prepSCA(){
	//prep chromophore absorption matrix
	float **chromMat = new float*[_lseantWlens];
	for (int i=0; i < _lseantWlens; i++){
		float wlen = _wlens[_startind + i];
		chromMat[i] = _chrom.getAbsArray(wlen);
		for (int j=0; j < _endmembers; j++){
			cerr << chromMat[i][j] << " ";
		}
		cerr << endl;
	}

	//calculate A^T * A
	float *ATA = new float[_endmembers*_endmembers];
	for (int i=0; i < _endmembers; i++){
		for (int j=0; j < _endmembers; j++){
			ATA[i*_endmembers + j] = 0;
		}
	}
	for (int i=0; i < _endmembers; i++){
		for (int k=0; k < _endmembers; k++){
			float temp = 0;
			for (int j=0; j < _lseantWlens; j++){
				temp += chromMat[j][i]*chromMat[j][k];
			}
			ATA[k*_endmembers + i] = temp;
		}	
	}
	cudaMalloc(&_GPUATA, sizeof(float)*_endmembers*_endmembers);
	cudaMemcpy(_GPUATA, ATA, sizeof(float)*_endmembers*_endmembers, cudaMemcpyHostToDevice);
	delete [] ATA;

	//transpose chromophore matrix and upload to GPU
	float *chromHost = new float[_endmembers*_samples];
	for (int i=0; i < _endmembers; i++){
		for (int j=0; j < _lseantWlens; j++){
			chromHost[i*_lseantWlens + j] = chromMat[j][i];
		}
	}
	cudaMalloc(&_GPUA, sizeof(float)*_endmembers*_samples);
	cudaMemcpy(_GPUA, chromHost, sizeof(float)*_endmembers*_samples, cudaMemcpyHostToDevice);

	//some other tjafs
	size_t pitch;
	cudaMallocPitch(&_gpuerror, &pitch, sizeof(float)*_threadsPerBlock, _samples/_threadsPerBlock);
	cudaMallocPitch(&_mu, &pitch, sizeof(float)*_threadsPerBlock, _endmembers*_samples/_threadsPerBlock);
	cudaMallocPitch(&_res, &pitch, sizeof(float)*_threadsPerBlock, _endmembers*_samples/_threadsPerBlock);

	delete [] chromHost;
	for (int i=0; i < _lseantWlens; i++){
		delete [] chromMat[i];
	}
	delete [] chromMat;
}

//do fitting using SCA
void SCAFitting::doFitting(GPUSkinData *skinData, float *gpuMua, bool shouldCopy, bool shouldUpdateSkindata, float *prevChrom, int prevMethbind){
	if (prevMethbind < 0){
		prevMethbind = 0;
	}
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	dim3 dimBlock(_threadsPerBlock);
	dim3 dimGrid(block_number_x, block_number_y);
	size_t inputPitch = _pitch/sizeof(float);
	const int MAX_ENDMEMBERS = 10;
	const int MAX_BLOCKDIM = 160;
	switch(_endmembers){
		case 2:
		SCA<2, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 3:
		SCA<3, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 4:
		SCAFast<4><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		//SCA<4, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 5:
		SCAFast<5><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		//SCA<5, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 6:
		SCA<6, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 7:
		SCAFast<7><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		//SCA<7, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 8:
		SCA<8, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 9:
		SCA<9, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		case 10:
		SCA<10, MAX_BLOCKDIM><<<dimGrid, dimBlock>>>(_GPUA, _GPUATA, gpuMua, _res, _mu, prevChrom, prevMethbind, _currMethbInd, _startind, _lseantWlens, _endmembers, inputPitch);
		break;
		default:
		cerr << "ERROR! This program was created in an infernally idiotic way. You need to add more cases in the switch statement in SCAFitting::doFitting, you tried to fit more parameters than was defined by the switch statement. (This is needed because of just in time compilation and the need for constants in the initialization of the CUDA kernel)" << endl;
		exit(1);
	}

	if (shouldCopy){
		block_number_y = _lseantWlens;
		dimGrid = dim3(block_number_x, block_number_y);
		removeMuam<<<dimGrid, dimBlock>>>(gpuMua, _res, _GPUA, _startind, _lseantWlens, _chrom.getNumEndmembers(), _chrom.getMelInd(), inputPitch);
	} 
	if (shouldUpdateSkindata){
		block_number_y = 1;
		dimGrid = dim3(block_number_x, block_number_y);
		calcOxyBvf<<<dimGrid, dimBlock>>>(_res, skinData->oxy, skinData->bvf, inputPitch);
		
	}
}

//calculate reflectance error after inversion
float* SCAFitting::reflerror(GPUOpticalProps *optProps, GPUSkinBase *skinBase, float *measrefl){
	int block_number_x = _samples/_threadsPerBlock;
	int block_number_y = 1;
	dim3 dimBlock(_threadsPerBlock);
	dim3 dimGrid(block_number_x, block_number_y);
	int startblockInd = _startind;
	size_t inputPitch = _pitch/sizeof(float);

	int reflerrorstartwlen = startblockInd;
	int reflerrorendwlen = _lseantWlens+_startind;
	int reflerrorantwlen = reflerrorendwlen-reflerrorstartwlen;
	int diff = -1*startblockInd+reflerrorstartwlen;

	ReflIsoL2ErrorCheck<<<dimGrid, dimBlock>>>(optProps->muae, optProps->muse, optProps->musd, skinBase->gcol, _GPUA, _res, _endmembers, reflerrorantwlen, measrefl, _gpuerror, inputPitch, reflerrorstartwlen, diff);
	return _gpuerror;
}




