//input: spectral image. Output: float image where each band corresponds to a chromophore
class GPUDM{ //omitted inherited classes
	public:
		GPUDM(int bands, int samples, const std::vector<float> &wlens);

		//
		//omitted framework functions
		//

		//inversion
		void invertReflectance(float *refl, bool printToFile, int pixel = -1); //takes in a host allocated array, uploads it to the GPU, inverts it using the isotropic diffusion model
		void invertMuam694(int iterations, bool printToFile, int pixel = -1); //clutterfunction for inverting muam694. Just to clean up execute() a bit
		void invertMuad(GPUSkinData *skinData, float *simrefl, float *deriv, float *currRefl, int startwlenInd, int endwlenInd, cudaStream_t *processingStream);
		void invertMuae(GPUSkinData *skinData, float *simrefl, float *deriv, float *currRefl, int startwlenInd, int endwlenInd, cudaStream_t *processingStream, bool shouldCalcSkinData);
		void detectMelaninType(float *refl, GPUSkinData *skinData); //detect correct melanin type based on the muad scaling at input wavelengths. Will only be able to determine when there is "pure" pheomelanin and "pure" eumelanin, but a mix is more likely. 
		void checkReflectanceScaling(float *refl, GPUSkinData *skinData, int wlenind1, int wlenind2, dim3 gridDim, size_t pitch);

		//spectral unmixing
		void doOneChromSpectralUnmixing(GPUSkinData *skinData, float *gpuMua, int startwlenInd, int endwlenInd, cudaStream_t *processingStream, MelaninDetermination choice);

	private:
		int num_melanin_iterations;

		//inversion arrays, all allocated on the GPU
		GPUSkinData *_skinDataMuadInv;
		GPUSkinData *_skinDataMuaeInv;
		GPUOpticalProps *_optProps;
		GPUSkinBase *_skinBase;
		float *_calarray;
		float *_simrefl;
		float *_deriv;
		float *_currRefl;
		int _numIterations;

		//image properties
		int _bands;
		int _samples; 
		std::vector<float> _wlens;
		bool _shouldCal;

		//GPU properties
		size_t _byteWidth;
		int _height;
		size_t _pitch;
		int _threadsPerBlock;
		dim3 _dimBlock;
		dim3 _dimGrid;
		cudaStream_t processingStream;
		cudaStream_t transferStream;

		//melanin fitting variables
		float _factor;
		float *_melcurve;
		float *_gpumeltype; //melanin type. 0: svaasand, 1: pheomelanin, 2: eumelanin
		float *_gpureflscaling; //reflectance scaling calculated from some wavelengths
		int _scalewlenind1;
		int _scalewlenind2;

		//framework properties
		//
		// omitted
		//

		//wavelength range for melanin determination	
		int _startwlenInd;
		int _endwlenInd;
		float _startwlen;
		float _endwlen;

		SCAFitting *sca450;
		SCAFitting *sca530;
		SCAFitting *sca700;
		SCAFitting *scabefore;

		//to ensure methb inclusion in first SCA fitting
		float *methbdummy;
		int methbDummyInd;
		
		//forward simulation
		float *_gpumuad;
		float *_gpures;
};


const float SCALE_WLEN_1 = 541;
const float SCALE_WLEN_2 = 576;


//fitting on the GPU using SCA
class SCAFitting : public LSEFitting{
	public:
		SCAFitting(Chromophores chrom, const std::vector<float> &wlens, float startwlen, float endwlen, int samples, int bands, int threadsPerBlock, size_t pitch);
		void doFitting(GPUSkinData *skinData, float *gpuMua, bool shouldCopy, bool shouldUpdateSkindata, float *prevChrom, int prevMethbind);
		float *getRes(){return _res;}; //return current chromophore fitting result
		int getNumEndmembers(){return _endmembers;};
		Chromophores getChrom(){return _chrom;};
		int getMethbInd(){return _currMethbInd;};
		float *reflerror(GPUOpticalProps *optProps, GPUSkinBase *skinBase, float *measrefl); //calculate the error from forward reflectance
		int getFitStartInd(){return _startind;};
		int getFitEndInd(){return _endind;};
	protected:
		void prepSCA(); //create arrays
		float *_GPUATA; //H matrix on the GPU
		float *_GPUA; //transpose of chromophore matrix on GPU
		float *_mu; //den mu-greia i algoritmen anae rikke hva den heter
		int _startind; //start index for fitting
		int _endind;
		int _currMethbInd;

		float *_gpuerror; //for forward reflectance calculation, not really a reasonable part of this class but whatever
};

//fitting using ISRA
class ISRAFitting : public SCAFitting{
	public:
		ISRAFitting(Chromophores chrom, const std::vector<float> &wlens, float startwlen, float endwlen, int samples, int bands, int threadsPerBlock, size_t pitch);
		void doFitting(GPUSkinData *skinData, float *gpuMua);
};
